import _ from 'lodash/fp';
import Immutable from 'immutable';

// The default value of every document.
const defaults = {
  data: null,
};

// The document type.
type Document = {
  data: any,
};

// The document constructor/factory.
const document = (data: Object = {}, base: Object = defaults): Document => (
  Immutable.Map.isMap(data)
    ? Immutable.Map(base).merge(data)
    : Immutable.Map(_.merge(base, data)));

// Convert a document to plain JavaScript.
const toJS = (m: Document): Object => m.toJS();

// The identity element of Document is a document with the default values.
const id = (): Document => document();

// Map a function over the contents of a document, the same as _.map.
// FIXME: How to annotate this construction??
const fmap = _.curry((f, m) => m.map(f));

// `mempty` is just an alias for the identity element.
const mempty = id;

// A binary operation that merges two documents.
const mappend = _.curry((m1, m2) => m1.merge(m2));

// Merge a list of documents. Can be either an `Immutable.List` or an `Array`.
const mconcat = (ms: Array<Document>): Document => (
  Immutable.List.isList(ms)
    ? ms.reduce((memo, m) => mappend(memo, m))
    : _.reduce((memo, m) => mappend(memo, m), id(), ms));

export default {
  document, toJS, defaults,
  id, fmap,
  mempty, mappend, mconcat,
};
